/*
 * This program and all associated contents are
 * Copyright (C) 1996-2005 David G. Andersen
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
*
 * In addition to the terms of the GPL, any web site featuring
 * this code in use as a CGI script must display the text
 * "The pi search engine is Copyright (C) 2001 David G. Andersen"
 * and provide a link to http://www.angio.net/pi/piquery
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */

#include <stdio.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "cgi-lib.h"

#define DO_CGI
#define DO_LOGFILE

#define MAX_OUTPUT_LEN 1000
#define LOGFILE "/home/dga/public_html/pi/pilog"
#define PIFILE "/home/dga/public_html/pi/pi200.4.bin"
/*#define PIFILE "/home/dga/classes/6.854/pi/pi50.4.bin"*/
#define MY_URL "http://www.angio.net/pi/bigpi.cgi"

#define HEADER "/home/dga/public_html/pi/pisearch.head.html"
#define FOOTER "/home/dga/public_html/pi/pisearch.foot.html"

typedef struct {
    char name[128];
    char val[128];
} entry_struct;

void
print_file(char *fname, char *procsec)
{
    FILE *f;
    char buf[BUFSIZ];

    f = fopen(fname, "r");
    if (!f) {
        printf("Error opening file: %s\n", strerror(errno));
        return;
    }
    while (fgets(buf, sizeof(buf), f)) {
	char *p;
	if ((p = strstr(buf, "PROCSEC")) != NULL) {
		*p = NULL;
        	puts(buf);
		puts(procsec);
		puts(p + 7);
	} else {
		puts(buf);
	}
    }
    fclose(f);
}

void
error(char *errmsg)
{
    printf("%s", errmsg);
    print_file(FOOTER, "0");
    exit(0);
}

char
*decode4(char *buf4, int pos, int len)
{
    char *ret;
    int i, dig;
    int j = 0;
    
    ret = malloc(len+1);
    if (!ret) {
	error("Could not allocate decode4 buffer\n");
    }

    for (i = pos; i < pos+len; i++) {
	dig = (unsigned char)buf4[i/2];
	if (!(i%2)) { /* First digit in the byte */
	    dig = (dig & 0xf0) >> 4;
	} else {
	    dig = (dig & 0x0f);
	}
	ret[j++] = dig+'0';
    }
    ret[len] = 0;
    return ret;
	
}

void
output4(char *buf4, int pos, int len)
{
    int i, dig;
    char r;

    pos -= 1;
    if (pos < 0) {
      return;
    }
    
    for (i = pos; i < pos+len; i++) {
	dig = (unsigned char)buf4[i/2];
	if (!(i%2)) { /* First digit in the byte */
	    dig = (dig & 0xf0) >> 4;
	} else {
	    dig = (dig & 0x0f);
	}
	r = dig+'0';
	putchar(r);
    }
}

size_t search(char *buf, char *searchstring, size_t filesize, size_t startpos);

int
setup(char *file, int *retfd, char **retbuf, size_t *fsize)
{
    int rc;
    int fd;
    char *buf;
    struct stat statbuf;
    size_t filesize;

    fd = open(file, O_RDONLY);
    if (fd < 0) {
	perror(file);
	return -1;
    }

    rc = fstat(fd, &statbuf);
    if (rc) {
	perror("could not stat file");
	return -1;
    }
    filesize = statbuf.st_size;

    buf = mmap(NULL, filesize, PROT_READ , MAP_SHARED, fd, 0);

    if (buf == MAP_FAILED) {
	perror("could not mmap file");
	return -1;
    }

#ifdef DO_MEMADVISE
    rc = madvise(buf, filesize, MADV_SEQUENTIAL);
#endif

    *retfd = fd;
    *retbuf = buf;
    *fsize = filesize;

    return 0;
}

long
timediff(struct timeval *time_now, struct timeval *time_prev)
{
        long usecs_now;
        
        if (time_prev->tv_sec == 0 &&
            time_prev->tv_usec == 0) {
                return 0;
        }
        
        usecs_now = (time_now->tv_sec - time_prev->tv_sec) * 1000000;
        usecs_now += time_now->tv_usec;

        return usecs_now - time_prev->tv_usec;
}


int
main(int argc, char **argv)
{
  int myfd;
  char *mybuf;
  char *searchkey = NULL;
  char *startpos = NULL;
  char *querytype = NULL;
  long int startposd = 0;
  int pos;
  int rc;
  size_t filesize;
  int i, j, keylen;
  LIST *head;

  struct timeval time_start, time_end;
  unsigned long ex_us;
  char procsec[16];
  
#ifdef DO_LOGFILE
  FILE *logfile;
#endif

  close(STDERR_FILENO);
  dup2(STDOUT_FILENO, STDERR_FILENO);

  gettimeofday(&time_start, NULL);
#ifdef DO_CGI

  head =  cgi_input_parse();

  printf("Content-type: text/html%c%c",10,10);
  print_file(HEADER, "");
  fflush(stdout);

  if (head == NULL) {
      error("<h2>You need to specify a request method string.</h2>\n");
  }

  searchkey = find_val(head, "UsrQuery");
  startpos = find_val(head, "startpos");
  querytype = find_val(head, "querytype");

#else
  searchkey = argv[1];
  if (argc > 2) {
    startpos = argv[2];
  }
  if (argc > 3) {
      querytype = argv[3];
  }
#endif

  if (searchkey == NULL || searchkey[0] == '\0') {
      error("<H2>You need to specify a search string.</H2>\n");
  }

  while (*searchkey && isspace(*searchkey)) { searchkey++; }

  /* Ignore punctuation, detect non-numeric searches */
  keylen = strlen(searchkey);
  for (i = j = 0; i < keylen; i++, j++) {
    if (searchkey[i] == ',' || searchkey[i] == '.') {
      j--;
      continue;
    } else if (!isdigit(searchkey[i])) {
      error("<H2>You can only search for digits (0-9), try again</h2>\n"
	    "As an example, the search \"pi\" is invalid, but \n"
	    "the search \"3232\" is perfectly OK.  Don't put in\n"
	    "the quote marks, of course\n");
    }
    else {
      searchkey[j] = searchkey[i];
    }
  }
  searchkey[j] = '\0';

  if (!startpos || !startpos[0] || !isdigit(startpos[0])) {
	startposd = 0;
  } else {
    int poslen = strlen(startpos);
    for (i = j = 0; i < poslen; i++, j++) {
      if (startpos[i] == '.' || startpos[i] == ',') {
	j--;
	continue;
      }
      startpos[j] = startpos[i];
    }
    startpos[j] = '\0';
    startposd = atoi(startpos);
  }

  rc = setup(PIFILE, &myfd, &mybuf, &filesize);

  if (rc) {
      error("An error occured setting up the file\n");
  }

  if (startposd > filesize*2) {
      error("You can't start searching for something after the end of the file..\n");
  }

#ifdef DO_LOGFILE
  /* XXX:  Add more to log */
  logfile = fopen(LOGFILE, "a");
  if (logfile != NULL) {
    fprintf(logfile, "%s %ld\n", searchkey, startposd);
    fclose(logfile);
  }
#endif

  if (querytype && !strcmp(querytype, "substr")) {
      int qlen = atoi(searchkey);
      if (startposd < 1) {
          printf("<H1>Bad query:  Must start at 1 or more</H1>\n");
      }
      else if (qlen > MAX_OUTPUT_LEN ||
	  startposd + qlen > (filesize*2)) {
	  printf("<H1>Bad query:  Too long</h1>\n");
      } else {
	  printf("<H2>Pi from %ld to %ld</h2>\n",
		 startposd, startposd+qlen-1);
	  output4(mybuf, startposd, qlen);
	  printf("<P>\n");
      }
      
  } else {
      pos = search(mybuf, searchkey, filesize, startposd);

      if (pos == -1) {
	  printf("<H2>The string %s did not occur in the first %d digits "
		 "of pi after position %ld.</H2>\n",
		 searchkey, filesize*2, startposd);
      }
      else {
	  char *tmpstr;
	  int i, l, ncom;
	  char withcom[32];

	  bzero(withcom, sizeof(withcom));
	  sprintf(withcom, "%d", pos+1);
	  l = strlen(withcom);
	  ncom = (l-1) / 3;
	  for (i = 0; i < l; i++) {
	    if (i > 0 && !(i % 3)) {
	      withcom[l - 1 + ncom - i] = ',';
	      ncom--;
	    }
	    withcom[l - 1 + ncom - i] = withcom[l - 1 - i];
	  }
	  	  
	  printf("The string <B>%s</b> was found at position %s "
		 "counting from the first digit after the decimal point."
		 "  The 3. is not counted.\n",
		 searchkey, withcom); fflush(stdout);
	  printf("<form method=\"post\" action=\"%s\">\n", MY_URL);
	  printf("<input type=\"hidden\" name=\"UsrQuery\" value=\"%s\">\n", searchkey);
	  printf("<input type=\"hidden\" name=\"startpos\" value=\"%d\">\n", pos+1);
	  printf("<input type=\"submit\" value=\"Find Next\">\n");
	  printf("</form>\n");
#if 0
	  printf("<P><B><A href=\"%s?UsrQuery=%s&startpos=%d\">Find Next</a></b>\n",
		 MY_URL, searchkey, pos+2);
#endif
      
	  printf("<P>The string and surrounding digits:</P><P>\n");

	  /* Surrounding digits */
	  if (pos > 20) {
	      tmpstr = decode4(mybuf, pos-20, 20);
	  } else {
	      tmpstr = decode4(mybuf, 0, pos);
	  }
	  printf("%s", tmpstr);
	  free(tmpstr);

	  /* The search key, decoded from the buffer so we can verify it */
	  tmpstr = decode4(mybuf, pos, strlen(searchkey));
	  printf(" <B>%s</B> ", tmpstr);
	  free(tmpstr);

	  /* The following digits */
	  if (pos+strlen(searchkey) < filesize*2-20) {
	      tmpstr = decode4(mybuf, pos+strlen(searchkey), 20);
	  } else {
	      tmpstr = decode4(mybuf, pos+strlen(searchkey),
			       filesize*2-pos-strlen(searchkey));
	  }
	  printf(" %s\n", tmpstr);
	  free(tmpstr);
      }
 }
  gettimeofday(&time_end, NULL);
  ex_us = timediff(&time_end, &time_start);
  sprintf(procsec, "%f", ((float)ex_us/1000000.0F));
#ifdef DO_CGI
  print_file(FOOTER, procsec);
#endif
  munmap(mybuf, filesize);
  exit(0);
}


